/*
 *     Botendo
 *     Copyright (C) 2023  limited_dev
 *
 *     This program is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */

package de.limited_dev.botendo.extensions.music.components

import de.limited_dev.botendo.util.Logger
import dev.schlaubi.lavakord.audio.TrackEndEvent
import dev.schlaubi.lavakord.audio.TrackExceptionEvent
import dev.schlaubi.lavakord.audio.TrackStuckEvent
import dev.schlaubi.lavakord.audio.on
import dev.schlaubi.lavakord.audio.player.Player
import dev.schlaubi.lavakord.rest.TrackResponse
import java.util.concurrent.BlockingQueue
import java.util.concurrent.LinkedBlockingQueue

class GuildTrackScheduler(val pl: Player) {

    private var queue: BlockingQueue<TrackResponse.PartialTrack> = LinkedBlockingQueue()
    var repeating = false
    private var hasRegisteredEvents = false

    ///Add a track to queue and start playing, if there is no song currently playing
    suspend fun queue(track: TrackResponse.PartialTrack) {
        if (this.pl.playingTrack == null) {
            play(track)
        } else {
            queue.offer(track)
        }
    }

    suspend fun playNext() {
        if (!queue.isEmpty())
            this.pl.playTrack(queue.poll())
        else
            this.pl.stopTrack()
    }

    private suspend fun play(tr: TrackResponse.PartialTrack) {
        this.pl.playTrack(tr)
    }

    fun addEvents() {
        if (hasRegisteredEvents)
            return
        hasRegisteredEvents = true

        Logger.out("Adding track events to GuildTrackScheduler...")

        pl.on<TrackEndEvent> {
            onTrackEnd(this)
        }

        pl.on<TrackStuckEvent> {
            onTrackStuck(this)
        }

        pl.on<TrackExceptionEvent> {
            onTrackExc(this)
        }
    }

    private suspend fun onTrackEnd(e: TrackEndEvent) {
        if (e.reason.mayStartNext) {
            if (repeating) {
                this.pl.playTrack(e.track.copy())
                return
            }
            Logger.out("Track has ended; Playing next...")
            playNext()
        }
    }

    private suspend fun onTrackStuck(e: TrackStuckEvent) {
        Logger.out("Track is stuck, retrying...")
        this.pl.playTrack(e.track.copy())
    }

    private suspend fun onTrackExc(e: TrackExceptionEvent) {
        Logger.out("Track had an exception, retrying...")
        this.pl.playTrack(e.track.copy())
    }

    fun clear() {
        Logger.out("Clearing queue...")
        repeating = false
        queue.clear()
    }

    fun getQueue(): List<TrackResponse.PartialTrack> {
        return queue.toList()
    }

    fun getHead(): TrackResponse.PartialTrack {
        return queue.first()
    }
}

